\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/MPL/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore FORTH.F 1.20
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\ Port to the gforth system and extensions by Ulrich.E.Hoffmann AT xlerb.de
\
\ Introduction  see: doc/uCore_Introduction.pdf
\ Tools         see: doc/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: doc/uCore_License.pdf
\ Conventions   see: doc/uCore_Conventions.pdf
\ Contributors  see: doc/uCore_Contributors.pdf
\
\ Standard Forth words composed of primitives
\

WITH_NOS [NOTIF]
: rot       ( n1 n2 n3 -- n2 n3 n1 )  >r over >r nip r> r> over >r nip r> ;
: -rot      ( n1 n2 n3 -- n3 n1 n2 )  over >r nip over >r nip r> r> ;
: tuck      ( n1 n2 -- n2 n1 n2 )     swap over ;
: under     ( n1 n2 -- n1 n1 n2 )     over swap ;
[THEN]

: (do    ( limit start -- count end )  under - 1- swap 1- ;
: (?do   ( limit start -- count end )  under -    swap 1- ;

Compiler
: DO     ( n1 n2 -- )  ( R: -- n3 n4 )                 ?comp T (do >r FOR H ;
: ?DO    ( n1 n2 -- )  ( R: -- n3 n4 )                 ?comp T (?do >r ?FOR H ;
: LOOP   ( -- )  ( R: n1 n2 -- n1 n2 )  ( R: n1 0 -- ) ?comp T NEXT rdrop H ;
Target

: I      ( -- n1 )  ( R: n2 n3 -- n2 n3 )   r>  r> r@ over - swap >r swap >r ;

Macros
Macro: abs    ( n -- u )       ?comp 2 lit, T ns-branch H 0 lit, T swap- H ;  \ neg IF negate THEN ;
Macro: bounds ( start len -- limit start )  ?comp T over >r + r> H ;
Target

: execute   ( addr -- )      >r ;
: 0<        ( n1 -- f )      dup neg IF  invert  THEN  xor ;
: <         ( n1 n2 -- f )        -   neg IF  drop true  EXIT THEN  drop false ;
: >         ( n1 n2 -- f )    swap-   neg IF  drop true  EXIT THEN  drop false ;
: >=        ( n1 n2 -- f )        -   neg IF  drop false EXIT THEN  drop true ;
: u<        ( n1 n2 -- f )        - carry IF  drop false EXIT THEN  drop true ;
: u>        ( n1 n2 -- f )    swap- carry IF  drop false EXIT THEN drop true ;
: max       ( n1 n2 -- max ) 2dup_-   neg IF  drop nip   EXIT THEN  2drop ;
: min       ( n1 n2 -- min ) 2dup_-   neg IF  2drop      EXIT THEN  drop nip ;
: umin      ( n1 n2 -- max ) 2dup_- carry IF  drop nip   EXIT THEN  2drop ;
: umax      ( n1 n2 -- min ) 2dup_- carry IF  2drop      EXIT THEN  drop nip ;
: within    ( w [low [high -- flag )   over - >r - r> - carry IF drop false EXIT THEN drop true ;
: extend    ( n -- d )       neg IF  -1  EXIT THEN  0 ;

WITH_TOR [IF]
: um*     ( u1 u2 -- udprod )   umultiply ;
: *       ( n1 n2 -- prod )     umultiply drop ; \ no overflow check!

: um/mod  ( uddividend udivisor -- urem uquot )   udivide ;
: u/      ( u1 u2 -- uquot )    0 swap udivide nip ;
: umod    ( u1 u2 -- urem )     0 swap udivide drop ;

: m/mod  ( d n -- mod quot )   
   dup >r   abs >r   neg IF  r@ +  THEN  r> um/mod   
   r> neg IF  >r negate over IF  swap r@ + swap 1-  THEN  r>  THEN  drop 
;
: /mod   ( n1 n2 -- rem quot ) >r extend r> m/mod ;
: /      ( n1 n2 -- quot )     /mod nip ;
: mod    ( n1 n2 -- rem )      /mod drop ;
[THEN]

: +!        ( n addr -- )      ld >r + r> ! ;
: increment ( addr -- )        ld >r 1+ r> ! ;
: decrement ( addr -- )        ld >r 1- r> ! ;
: on        ( addr -- )        -1 swap ! ;
: off       ( addr -- )         0 swap ! ;

: >>        ( val u -- val' )  ?FOR u2/ NEXT ;
: <<        ( val u -- val' )  ?FOR 2* NEXT ;
: shift     ( val n -- val' )  neg IF  abs >>            EXIT THEN << ;
: ashift    ( van n -- val' )  neg IF  abs ?FOR 2/ NEXT  EXIT THEN << ;
: clear     ( ??? -- ) 0 dsp! ;
: case?     ( n1 n2 -- n1 ff | tf ) \ Selection Operator. n1 eingaengiger Code. n2 Compare Code
                                \ wenn die beide gleich sind, dann wird Trueflag hinterlaessen,
                                \ sonst Falseflag. Code n1 wird nicht gendert und bleibt als
                                \ zweites Element auf dem Stack liegen.
   over - 0= zero not IF  nip  THEN
;
: fill     ( addr len pattern -- )  swap >r swap 1- r> ?FOR  over swap 1 +st  NEXT  2drop ;
: cmove ( addr1 addr2 n -- ) \ from addr1 to addr2 n times
   >r 1- swap 1- r> ?FOR  1 +ld >r swap 1 +st r>  NEXT 2drop
;
: count ( addr -- addr+1 n ) ld 1+ swap ;
: d+  ( d1 d2 -- d3 )  >r swap >r +  r> r> +c ;
: d-  ( d1 d2 -- d3 )  >r swap >r -  r> r> invert +c ;
: dnegate ( d -- ~d ) invert swap invert swap 1 0 d+ ;
: dabs    ( d -- +d ) neg IF dnegate THEN ;
: 2swap ( d1 d2 -- d2 d1 )  rot >r rot r> ;
: 2over ( d1 d2 -- d1 d2 d1 ) >r over r> swap >r >r >r over r> swap r> swap r> ;
: 2!    ( d addr -- )  st 1 +st drop ;
: 2@    ( addr -- d )  1 +ld -1 +ld drop ;
: ei       ( -- )    status@ #ie or status! ;
: di       ( -- )    status@ [ #ie invert ] Literal and status! ;
