\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/License/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore GFORTH-CONFIG.F 1.20
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\ Port to the gforth system and extensions by Ulrich.E.Hoffmann AT xlerb.de 
\
\ Introduction  see: doc/uCore_Introduction.pdf
\ Tools         see: doc/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: doc/uCore_License.pdf
\ Conventions   see: doc/uCore_Conventions.pdf
\ Contributors  see: doc/uCore_Contributors.pdf
\
\ MicroCore Forth Cross-compiler configuration

Create _GFORTH

warnings off

: mem-save ( x1 ... xn n -- addr ) 
     dup 1+ cells allocate Abort" memstore: Cannot allocate memory"  ( x1 ... xn n addr )
     dup >r
     2dup ! cell+  \ store length 
     swap 0 ?DO ( x1 .. xi addr_i )
         swap over ! cell+  ( x1 ... xi-1 addr_i-1 )
     LOOP ( addr1 )
     drop r> ;

: mem-restore ( addr -- x1 ... xn n )
    dup >r ( addr )
    dup @  ( addr len )
    dup >r cells +  ( addr' )
    r@ 0 ?DO ( x1 .. xi addri )
       dup @ swap 1 cells -   ( x1 ... xi xi+1 addr_i+1 ) 
    LOOP ( x1 ... xn addr )
    drop r> ( x1 ... xn n )  
    r> free Abort" mem-restore: could not free memory" ;

: source> ( -- sca )  save-input mem-save  ;
: >source ( sca -- ) mem-restore  restore-input abort" couldn't restore input" ;
: drop-source ( sca -- )  free Abort" drop-source: cannot free source code info" ;

: u2/ ( u1 -- u2 )   1 rshift ; \ Assumes 2's complement arithmetic
: binary             2 base ! ;
: cell-              1 cells - ;

\ IO Redirection
\ make it similar to Win32Forth's IO redirection using _CR _EMIT _TYPE

: _CR ( -- )  newline (type)  ;
Defer CR  

: _EMIT ( c -- ) (emit) ;

: _TYPE ( c-addr len -- )
    2dup newline d= IF  2drop  CR EXIT THEN  \ make old cr invoke new CR
    (type) ;

' _TYPE IS TYPE
' _CR is CR
' _EMIT is EMIT

: Module ( <spaces>ccc<spaces> -- )
   >IN @ >R  BL WORD FIND IF  EXECUTE  ELSE DROP THEN
   R> >IN !  MARKER ;

: ?comp   state @ 0= Abort" compilation only" ;
: ?exec   state @    Abort" execution only" ;

Variable theInterpreter   ' interpreter theInterpreter !
Variable theCompiler      ' compiler    theCompiler !

: [   postpone [   theInterpreter @  is parser ; immediate
: ]   ]    theCompiler @ is parser ;

: forget  ' drop ;

\ floored division
: /  ( n1 n2 -- n3  ) >r s>d r> fm/mod nip ;

: ?missing ( f -- ) ABORT" not found" ;

: defined ( name -- addr | 0 )
  parse-word name-too-short? find-name dup IF  name?int  THEN 
; 
