\ Do not use this file except in compliance with the License. You may
\ obtain a copy of the License at http://www.microcore.org/License/
\ Software distributed under the License is distributed on an "AS IS" basis,
\ WITHOUT WARRANTY OF ANY KIND, either express or implied.
\ See the License for the specific language governing rights and limitations
\ under the License.
\
\ The Original Code is: MicroCore MICROCORE.F 1.20
\
\ The Initial Developer of the Original Code is Klaus.Schleisiek AT hamburg.de.
\ Port to the gforth system and extensions by Ulrich.E.Hoffmann AT xlerb.de 
\
\ Introduction  see: doc/uCore_Introduction.pdf
\ Tools         see: doc/uCore_Implementation.pdf
\ Versions      see: uCross_Versions.txt
\ License       see: doc/uCore_License.pdf
\ Conventions   see: doc/uCore_Conventions.pdf
\ Contributors  see: doc/uCore_Contributors.pdf
\
\ MicroCore Forth Cross-compiler
\ This code loads on the public domain Win32Forth http://www.forth.org/
\ This code loads on the public domain gforth     
\                               http://www.gnu.org/software/gforth/gforth.html
\ Known Limitations:
\ V_1.10 can at most handle 30 bit literals in the target

Forth definitions decimal

: .version  ( -- )  ." 1.20 " ;
: .date     ( -- )  ." 24-JUN-2004 " ;

: 2**  ( n -- 2**n )   1 swap 0 ?DO  2*  LOOP ;

#prog_addr_width 2** Constant #maxprog
#data 2** 1-         Constant #progmask
#data 1- 2** negate  Constant #smallest

8                    Constant #code     \ CONSTANTS.VHD >> When changed, #literal has to be adapted as well <<
#code 2** 1-         Constant #codemask

#data #code 1- /mod swap 0= 1+ + Constant #nibbles  \ nibbles needed to represent any number
\
\ Some System word (re)definitions for a more sympathetic environment
\
include config.f

\
\ accessing the targets program memory
\
Create Memory   #progmask #maxprog umin 1+ cells allot

Variable Tcp                        \ TargetCodePointer
Variable Tdp                        \ TargetDataPointer
Variable Sequential                 \ number of past opcodes available for peep-hole optimization

Variable Macro      Macro off       \ holds the current macro tag to be included in code compilation. 0 => no macro
Variable Macro#     Macro# off      \ holds the last assigned macro number
Variable Macros     Macros off      \ linked list of all macros
$FF000000 Constant #macromask       \ macro# is embedded in opcode field

Variable Labels     Labels off      \ linked list of all labels
Variable Colons     Colons off      \ Builds a linked list of :-definitions for the disassembler
Variable Operators  Operators off   \ Linked list of all operators, used by the disassembler
Variable Constants  Constants off   \ linked list of all constants
Variable Variables  Variables off   \ linked list of all creates and variables
Variable Codes      Codes off       \ linked list of all threaded Code definitions

Variable TaskVariables TaskVariables off \ counter for task variables


: >memory  ( caddr -- addr )   cells Memory + ;
: t!       ( n caddr -- )      >memory ! ;
: t@       ( caddr -- n )      >memory @ ;
: opcode@  ( caddr -- op )     t@ #codemask and ;
: macro@   ( caddr -- # )      t@ -$18 shift ;
: macro!   ( n caddr -- )      swap Macro @ $18 shift or swap t! ;
: there    ( -- addr )         Tcp @ ;
: tallot   ( n -- )            dup Sequential +!   Tcp +! ;
: t,       ( n -- )
   dup [ #codemask invert ] Literal and abort" not an opcode"
   there macro!  1 Tcp +!  1 Sequential +!
;
: trap-vector ( n -- addr )   #usr_vect_width 2** * ;

: .addr   ( caddr -- )   Base @ >r  hex
   0 <#  [ #data 3 + 4 / ] Literal 0 DO  #  LOOP  #> type
   r> Base !
;
: .opcode ( opcode -- )  Base @ >r  hex  
   0 <#  [ #code 3 + 4 / ] Literal 0 DO  #  LOOP  #> type space
   r> Base !
;
\
\ peep-hole optimizer
\
: |        ( -- )              Sequential off ;             \ break peep-hole optimizer

: prev@    ( -- opcode | 0 )   Sequential @     IF  there 1 - t@  EXIT THEN  false ;

: pprev@   ( -- opcode | 0 )   Sequential @ 1 > IF  there 2 - t@  EXIT THEN  false ;

: Match  ( opcode mask -- )    Create  , ,
Does>  ( opcode -- f )  dup @ >r   cell+ @ xor r> and 0= ;

\
\ instruction fields as defined by the hardware architecture
\
$80 Constant #literal
$7F Constant #litmask
$1F Constant #usrmask
$60 Constant #op
$18 Constant #stack
$07 Constant #group

$00 Constant #bra
$20 Constant #alu
$40 Constant #mem
$60 Constant #usr

$00 Constant #none
$08 Constant #pop
$10 Constant #push
$18 Constant #both

\ binary Operators
  0 Constant #add
  1 Constant #adc
  2 Constant #sub
  3 Constant #ssub
  4 Constant #and
  5 Constant #or
  6 Constant #xor
  7 Constant #nos

\ complex Math Operators

  0 Constant #mults
  1 Constant #0divs
  2 Constant #udivs
\  3 Constant #divs
  4 Constant #ldivs
\  5 Constant #sqrts
\  6 Constant #crcs
  7 Constant #swap

\ unary operators
  0 Constant #not
  1 Constant #sl
  2 Constant #asr
  3 Constant #lsr
  4 Constant #ror
  5 Constant #rol
  6 Constant #zequ
  7 Constant #cc

\ conditions
  0 Constant #never
  1 Constant #always
  2 Constant #zero
  3 Constant #nzero
  4 Constant #sign
  5 Constant #nsign
  6 Constant #novl
  7 Constant #ncarry

\ branches
  0 Constant #dup
  1 Constant #exc
  2 Constant #?dup
  3 Constant #?ovl
  4 Constant #int
  5 Constant #iret
  6 Constant #thread
  7 Constant #token

\ registers
  0 Constant #status
  1 Constant #tor
  2 Constant #rstack
  3 Constant #local
  4 Constant #rsp
  5 Constant #dsp
  6 Constant #task
  7 Constant #ip

\
\ Opcode-field compilers and Opcode compiler
\
: Type:  ( <name> n -- )
   Create , 
Does> ( u1 u2 -- opcode )
   3 depth > abort" not enough fields specified"
   @ over #stack invert  and abort" not a stack field"  or
   over #group invert  and abort" not a group field"    or
;      

: don't   ( -- )  True abort" can not be executed" ;

: Field:  ( <name> n -- )  Constant 
;

: Group:  ( <name> n -- )  Constant 
;

: Op:  ( n <name> -- )
   Create ,                 \ here the instruction code is layed down which will be compiled into the target memory
   also Forth ' , previous  \ here the cfa of the word is layed down which will be executed during interpretation in the target
   here  Operators @ , Operators !
Does> ( -- )  exec? IF  cell+ @ execute  EXIT THEN  @ t, ;


1414 Constant #macro

: (Macro: ( <name> -- addr ) \   | xt | macro# | link |
  Create here 0 ,   1 Macro# +!   Macro# @ ,  here Macros @ , Macros !   
  Does>  dup cell+ @ Macro !  @ execute  Macro off
;

: Macro: ( <name> -- addr )   (Macro: :noname #macro ;

: ;macro ( addr xt #macro -- )
    #macro ?pairs
    postpone ;  ( addr xt )
    swap !
; immediate

\ redefine ; so that it can be used to terminate macro definitions
: ;  ( # -- ) 
   dup #macro = IF postpone ;macro EXIT THEN
   postpone ;
; immediate

\
\ The MicroCore "Assembler"
\
Vocabulary Target                \ Vocabulary for all Target definitions

: H  ( -- )  Forth  ; immediate  \ "Host" - short-hand definition to switch context also within a definition
: T  ( -- )  Target ; immediate

Target definitions

Vocabulary Assembler

: {  ( -- )   Assembler ; immediate     \ The braces allow to "delve" into the assembler during target compilation

: Asm  ( -- )   T Assembler H ;

Forth definitions

: Asm  ( -- )   T Assembler H ; immediate

Target Assembler definitions

: }  ( -- )   Target ; immediate        \ This switches back into "Forth"-mode

#alu Type: ALU
#bra Type: BRA
#mem Type: MEM
#usr Type: USR

#none Field: NONE
#push Field: PUSH
#pop  Field: POP
#both Field: BOTH

\ Binary operators
#add   Group: ADD
#sub   Group: SUB
#adc   Group: ADC
#ssub  Group: SSUB
#and   Group: AND
#or    Group: OR
#xor   Group: XOR
#nos   Group: NOS

\ complex Math Operators

#mults Group: MULTS
#0divs Group: 0DIVS
#udivs Group: UDIVS
\ #divs  Group: DIVS
#ldivs Group: LDIVS
\ #sqrts Group: SQRTS
\ #crcs  Group: CRCS
#swap  Group: SWAPS

\ Unary operators
#not   Group: NOT
#sl    Group: SL
#asr   Group: ASR
#lsr   Group: LSR
#ror   Group: ROR
#rol   Group: ROL
#zequ  Group: ZEQU
#cc    Group: CC

\ Conditions
#never  Group: NEVER
#always Group: ALWAYS
#zero   Group: ZERO
#nzero  Group: NZERO
#sign   Group: SIGN
#nsign  Group: NSIGN
#novl   Group: NOVL
#ncarry Group: NCARRY

\ Branches
#dup    Group: opDUP
#exc    Group: EXC
#?dup   Group: op?DUP
#?ovl   Group: ?OVL
#int    Group: INT
#iret   Group: IRET
#thread Group: THREAD
#token  Group: TOKEN

\ Register
#status  Group: STATUS
#tor     Group: TOR
#rstack  Group: RSTACK
#local   Group: LOCAL
#rsp     Group: RSP
#dsp     Group: DSP
#task    Group: TASK
#ip      Group: IP

\ offsets for post-incremeting @ and ! instructions
 0 Group: 0
 1 Group: 1
 2 Group: 2
 3 Group: 3
 4 Group: -4
 5 Group: -3
 6 Group: -2
 7 Group: -1

\
\ Target literals and addresses are composed of literal-nibbles
\ preceeding opcodes which use them
\
Target definitions Assembler

NEVER  NONE BRA Op: nop     noop  ( -- )

Forth definitions

: /nibble  ( lit -- nib lit' )
   dup #litmask and #literal or   swap #literal /
;
: last_nibble?  ( nib lit -- f )
   0 case? IF  [ #literal u2/ ] Literal and 0=  EXIT THEN
   -1    = IF  [ #literal u2/ ] Literal and     EXIT THEN
   drop False
;
: >nibbles  ( lit -- nib_q .. nib_1 q )  \ convert literal into q nibbles
   0 >r  BEGIN  r> 1+ >r   /nibble  2dup last_nibble? UNTIL  drop r>
;
: nibbles  ( lit -- quan )  \ number of nibbles needed to represent lit
   0 >r  BEGIN  r> 1+ >r   /nibble  tuck last_nibble? UNTIL  drop r>
;
: *nibble  ( nib lit' -- lit )
   #literal *   swap #litmask and or
;
: nibbles>  ( nib_q .. nib_1 q -- lit )  \ convert q nibbles into literal
   dup 0= ABORT" no nibbles available"
   >r  #litmask and
   dup [ #literal u2/ ] Literal and IF  #literal negate or  THEN
   r> 1 ?DO  *nibble  LOOP
;
: nibbles!  ( nib_q .. nib_1 quan caddr -- )  swap bounds DO  I macro!  LOOP ;

: prevlit? ( -- f )   prev@ #literal and ;

: ?nop,  ( -- )      prevlit? 0=EXIT
   Macro @   Macro off  T nop H   Macro !
;
: nibbles,  ( nib_q .. nib_1 quan -- )
   ?nop, there over tallot nibbles!
;
: nibbles@  ( caddr -- nib_q .. nib_1 quan )
   dup  BEGIN  dup t@ #literal and WHILE  dup t@ >r  1+  REPEAT
   swap -   dup  BEGIN  ?dup WHILE  1-  r> -rot  REPEAT
;
: ?literal ( lit -- lit' )   \ range check and sign extension
   dup #smallest and over 0< IF  #smallest xor  THEN  abort" number beyond data path width"
   dup #smallest and IF  #smallest or  ELSE  [ #smallest invert ] Literal and  THEN  \ sign extension
;
: lit, ( n --  | n )   \ compiles a literal into the minimum number of nibbles, leave on stack if interpreting
   ?literal exec? ?EXIT >nibbles nibbles,
;
: fixed,  ( n quan -- )      \ compiles a literal into quan nibbles.
   ?literal dup >r 0 DO  /nibble  LOOP  drop  r> nibbles,
;

DEBUGGING [IF]

: .lit  ( caddr -- )   nibbles@ nibbles> . ;

: tdump  ( caddr quan -- )
   cells bounds ?DO  cr I .addr ." :  "
                     I >memory $10 cells bounds DO  I @ .opcode  1 cells +LOOP
                $10 +LOOP ;

[THEN]

: prevsmall?  ( -- ff )
   prevlit? 0= abort" needs literal"   
   prev@ #litmask and [ #group u2/ invert ] Literal and   
   dup IF  [ #group u2/ 1+ ] Literal + #litmask and   THEN     \ either %0..00000 and %1..111-- is ok
;
\
\ Opcode definitions
\
Target definitions Assembler

\ Conditions
\ NEVER  NONE BRA Op: nop     noop  ( -- )
ALWAYS NONE BRA Op: exit         don't
ZERO   NONE BRA Op: z-exit       don't
NZERO  NONE BRA Op: nz-exit      don't
SIGN   NONE BRA Op: s-exit       don't
NSIGN  NONE BRA Op: ns-exit      don't
NOVL   NONE BRA Op: no-exit      don't
NCARRY NONE BRA Op: nc-exit      don't

NEVER  POP  BRA Op: drop_flag    don't ( n lit -- lit ) \ upholds lit status, used before z-branch
ALWAYS POP  BRA Op: branch       don't
ZERO   POP  BRA Op: z-branch     don't
NZERO  POP  BRA Op: nz-branch    don't
SIGN   POP  BRA Op: s-branch     don't
NSIGN  POP  BRA Op: ns-branch    don't
NOVL   POP  BRA Op: no-branch    don't
NCARRY POP  BRA Op: nc-branch    don't

NEVER  BOTH BRA Op: drop         drop  ( n -- )
ALWAYS BOTH BRA Op: call         don't
ZERO   BOTH BRA Op: z-call       don't
NZERO  BOTH BRA Op: nz-call      don't
SIGN   BOTH BRA Op: s-call       don't
NSIGN  BOTH BRA Op: ns-call      don't
NOVL   BOTH BRA Op: no-call       don't
NCARRY BOTH BRA Op: nc-call      don't

\ Branches
opDUP  PUSH BRA Op: dup          dup   ( n -- n n )
\ EXC    PUSH BRA Op: exc          don't
op?DUP PUSH BRA Op: ?dup         ?dup  ( n -- n n | 0 )
?OVL   PUSH BRA Op: ?ovl         don't
\ INT    PUSH BRA Op: int          don't
IRET   PUSH BRA Op: iret         don't
THREAD PUSH BRA Op: thread       don't
TOKEN  PUSH BRA Op: token        don't

\ Binary Operators
ADD    POP  ALU Op: +            +     ( n1 n2 -- n3 )
ADC    POP  ALU Op: +c           don't
SUB    POP  ALU Op: -            -     ( n1 n2 -- n3 )
SSUB   POP  ALU Op: swap-        don't
AND    POP  ALU Op: and          and
OR     POP  ALU Op: or           or
XOR    POP  ALU Op: xor          xor
NOS    POP  ALU Op: nip          nip

ADD    PUSH ALU Op: 2dup_+       don't ( n1 n2 -- n1 n2 n3 )
SUB    PUSH ALU Op: 2dup_-       don't ( n1 n2 -- n1 n2 n3 )
\ 2DUP or OVER OVER optimizer
NOS    PUSH ALU Op: over         over  ( n1 n2 -- n1 n2 n1 )

NOT    BOTH ALU Op: invert       invert
SL     BOTH ALU Op: 2*           2*
ASR    BOTH ALU Op: 2/           2/
LSR    BOTH ALU Op: u2/          u2/
ROR    BOTH ALU Op: ror          don't
ROL    BOTH ALU Op: rol          don't
ZEQU   BOTH ALU Op: 0=           0=
CC     BOTH ALU Op: cc           don't

\ Data Memory access
0      PUSH MEM Op: ld           don't
0      POP  MEM Op: st           don't

\ Registers
STATUS BOTH MEM Op: status@      don't
TOR    BOTH MEM Op: r@           don't ( -- n )
RSTACK BOTH MEM Op: r>           don't ( -- n )
LOCAL  BOTH MEM Op: +lld         don't
RSP    BOTH MEM Op: rsp@         don't
DSP    BOTH MEM Op: dsp@         don't
TASK   BOTH MEM Op: +tld         don't
IP     BOTH MEM Op: ip@          don't

STATUS NONE MEM Op: status!      don't
TOR    NONE MEM Op: r!           don't ( n -- )
RSTACK NONE MEM Op: >r           don't ( n -- )
LOCAL  NONE MEM Op: +lst         don't
RSP    NONE MEM Op: rsp!         don't
DSP    NONE MEM Op: dsp!         don't
TASK   NONE MEM Op: +tst         don't
IP     NONE MEM Op: ip!          don't

WITH_NOS [IF]
T Assembler SWAPS NONE ALU Op: swap swap  ( n1 n2 -- n2 n1 ) H
Macro: rot     ( n1 n2 n3 -- n2 n3 n1 )  comp? IF  T >r swap r> swap H EXIT THEN  rot ;
Macro: -rot    ( n1 n2 n3 -- n3 n1 n2 )  comp? IF  T swap >r swap r> H EXIT THEN  -rot ;
Macro: tuck    ( n1 n2 -- n2 n1 n2 )     comp? IF  T swap over       H EXIT THEN  swap over ;
Macro: under   ( n1 n2 -- n1 n1 n2 )     comp? IF  T over swap       H EXIT THEN  over swap ;
[ELSE]
Macro: swap    ( n1 n2 -- n2 n1 )        comp? IF  T over >r nip r>  H EXIT THEN  swap ;
[THEN]

WITH_TOR [IF]
T Assembler MULTS NONE ALU Op: mults don't H
Macro: umultiply ( mult1 mult2 -- prod_l prod_h )
        ?comp T >r H 0 lit, #data 0 ?DO  T mults H LOOP T nip r> H ;

T Assembler 0DIVS NONE ALU Op: 0divs don't
            UDIVS NONE ALU Op: udivs don't
            LDIVS NONE ALU Op: ldivs don't H
Macro: udivide ( div_l div_h divisor -- rem quot )
        ?comp T 0divs H #data 0 ?DO  T udivs H LOOP T ldivs nip r> H ;
[ELSE]
T Assembler SUB NONE ALU Op: over_swap_-  don't ( n1 n2 -- n1 n3 )
[THEN]

Macro: nor@     ( -- n )         ?comp T r> r@ over >r nip H ;
Macro: @        ( addr -- n )    ?comp T ld drop H ;
Macro: !        ( n addr -- )    ?comp T st drop H ;
Macro: l@       ( rel -- n )     ?comp T +lld drop H ;
Macro: l!       ( n rel -- )     ?comp T +lst drop H ;
Macro: t@       ( rel -- n )     ?comp T +tld drop H ;
Macro: t!       ( n rel -- )     ?comp T +tst drop H ;
     : +ld      ( n -- )         ?comp prevsmall? IF  T + ld H  EXIT THEN
        prev@ #group and [ Asm  0 PUSH MEM H ] Literal or  -1 tallot t,
     ;
     : +st     ( n -- )          
        ?comp prevsmall? IF  T + st H  EXIT THEN
        prev@ #group and [ Asm  0 POP MEM H ] Literal or  -1 tallot t,
     ;

-1 Constant FLAGS_REG
Macro: ie!     ( n -- )         ?comp -1 lit, T st drop H ;
Macro: flags@  ( -- n )         ?comp -1 lit, T ld drop H ;

-2 Constant TASK_REG
Macro: task@   ( -- addr )      ?comp -2 lit, T ld drop H ;
Macro: task!   ( addr -- )      ?comp -2 lit, T st drop H ;

WITH_TOKENS [IF] Macro: ip-branch ( -- )  ?comp T thread token H ;
          [ELSE] Macro: ip-branch ( -- )  ?comp T thread >r H ;
          [THEN]
Macro: 0=branch  ( f -- )         ?comp T drop_flag z-branch  H ;
WITH_TOR [IF] [ELSE]
Macro: 0<>branch ( f -- )         ?comp T drop_flag nz-branch H ;
[THEN]
Macro: rdrop     ( -- )           ?comp T r>  drop H ;
Macro: 2drop     ( n n -- )       comp? IF  T drop drop       H EXIT THEN  2drop ;
Macro: 2dup      ( d -- d d )     comp? IF  T over over       H EXIT THEN  2dup ;
Macro: c=0       ( -- )           ?comp  0 lit, T 2* drop H ;
Macro: c=1       ( -- )           ?comp -1 lit, T 2* drop H ;
Macro: =         ( n1 n2 -- f )   comp? IF  T - 0=            H EXIT THEN  = ;
Macro: negate    ( n -- -n )      comp? IF  0 lit, T swap-    H EXIT THEN  negate ;
     : Literal   ( -- n )         lit, ;
Macro: 1+        ( n1 -- n2 )     comp? IF   1 lit, T +       H EXIT THEN  1+ ;
Macro: 1-        ( n1 -- n2 )     comp? IF  -1 lit, T +       H EXIT THEN  1- ;
Macro: True      ( -- tf )       -1 lit, ;
Macro: False     ( -- ff )        0 lit, ;
Macro: cell+     ( n1 -- n2 )     comp? IF   1 lit, T +           H EXIT THEN  1+ ;
Macro: cell-     ( n1 -- n2 )     comp? IF  -1 lit, T +           H EXIT THEN  1- ;

\
\ branching
\ Due to relative addressing, the number of nibbles needed for a branch address
\ varies. At first, only one nibble is assumed to be sufficient to hold the offset.
\ When the compiler finds that more nibbles are needed, the source code is re-interpreted
\

Forth definitions

Variable Prefix     Prefix off     \ Prefix codes for conditionals to handle e.g. "carry IF" etc.

 0 Constant #if-bit
 1 Constant #nif-bit
 2 Constant #z-bit
 3 Constant #nz-bit
 4 Constant #s-bit
 5 Constant #ns-bit
 6 Constant #c-bit
 7 Constant #nc-bit
 8 Constant #o-bit
 9 Constant #no-bit
10 Constant #call
12 Constant #branch

: no_prefix       ( -- )    #if-bit Prefix ! ;

: ?strange_prefix ( n -- )  ?dup 0=EXIT no_prefix dup . abort" unexpected Prefix" ;

: >branch  ( caddr -- br.addr quan.nibbles )  \ length of branch lit field
   0  BEGIN  over t@ #literal and WHILE  1+ swap  1+ swap  REPEAT
;
: <mark    ( -- sca caddr )    ?nop, |  source> there ;

: <resolve ( sca caddr prefix -- )  \ ProgramCounter points behind branch opcode
   ?nop, |   1 >r                   \ minimum length of branch
   BEGIN  over there r@ + 1+ -      \ branch offset
          over  #if-bit = IF  1-  THEN
          over #nif-bit = IF  1-  THEN
          nibbles r@ >              \ does branch offset fit?
   WHILE  r> 1+ >r                  \ try with offset one nibble longer
   REPEAT
   dup #if-bit = swap #nif-bit = or IF  1-  THEN
   there r@ + 1+ -  r> fixed,  drop
;
: >mark    ( quan.nibbles prefix -- sca caddr )
   ?nop, |   swap >r   source> there
   rot dup #if-bit = swap #nif-bit = or IF  [ #literal 1 + ] Literal  ELSE  #literal  THEN
   r> 0 ?DO  dup t,  LOOP  drop
;
: >resolve ( sca caddr -- quan.nibbles.needed | 0 )
   ?nop, |  dup >r  dup >branch >r  \ R: offset width allocated
   swap t@ #litmask and +           \ fetch numerical value of first emplaced branch offset nibble and correct branch start address
   there swap 1+ - >nibbles         \ branch offset
   dup r> -                         \ does offset fit?
   IF  dup >r ndrop                 \ remove nibbles
       r> r> Tcp !  swap >source  EXIT
   THEN
   r> nibbles!  drop-source  false
;
: conditional,  ( -- )  Prefix @   no_prefix
   #if-bit  case? IF  T 0=branch   H  EXIT THEN   \ standard Forth consumes the flag
WITH_TOR [IF] [ELSE]
   #nif-bit case? IF  T 0<>branch  H  EXIT THEN
[THEN]
   #z-bit   case? IF  T z-branch   H  EXIT THEN
   #nz-bit  case? IF  T nz-branch  H  EXIT THEN
   #s-bit   case? IF  T s-branch   H  EXIT THEN
   #ns-bit  case? IF  T ns-branch  H  EXIT THEN
   #no-bit  case? IF  T no-branch  H  EXIT THEN
   #nc-bit  case? IF  T nc-branch  H  EXIT THEN
   ?strange_prefix
;

$101 Constant #if
$202 Constant #else
$303 Constant #begin
$404 Constant #while
$505 Constant #for
$606 Constant #colon
$707 Constant #trap
$808 Constant #?for
$909 Constant #code

: (if   ( quan.offset -- sca caddr # )
\   Prefix @ #c-bit = IF  T cc H #nc-bit Prefix !  THEN   \ handles carry IF case automatically, changing carry!
   Prefix @ >mark conditional, #if
;
: (else ( quan.offset -- sca caddr # )   #branch >mark T branch H #else ;

\ Due to the variable length branch offset fields ELSE does not compile anything
\ This is handled by THEN instead. ELSE leaves its unique check digit. Therefore,
\ THEN "knows" when it was preceeded by an ELSE.
\
\ THEN tests the check digit: If it was #ELSE,
\ the branch between ELSE ... THEN is compiled first yielding the actual branch
\ offset needed for ELSE. Thereafter, the IF ... ELSE branch is compiled.

Asm NEVER  NONE BRA H #codemask Match nop?
Asm NEVER  PUSH BRA H #codemask Match dup?
Asm ALWAYS BOTH BRA H #codemask Match call?

Target definitions Forth

: carry ( -- )    ?comp #nc-bit Prefix ! ;
: neg   ( -- )    ?comp #ns-bit Prefix ! ;
: ovl   ( -- )    ?comp #no-bit Prefix ! ;
: zero  ( -- )    ?comp #nz-bit Prefix ! ;
: not   ( -- )    ?comp Prefix @ 1 xor Prefix ! ;

: IF    ( -- sca caddr # )  ?comp 1 (if ;

: ELSE  ( sca1 caddr1 #1 -- sca1 caddr1 sca2 caddr2 #2 )
   ?comp #if ?pairs  1 (else
;
: THEN ( sca caddr #if | sca1 caddr1 sca2 caddr2 #else -- )
   ?comp #if case? IF  >resolve ?dup 0=EXIT  (if  EXIT THEN
         #else ?pairs
   dup >r  >resolve ?dup IF  rdrop (else  EXIT THEN
   r>   Tcp @ >r   >branch drop 1+ Tcp !
   >resolve ?dup IF  rdrop (if  EXIT THEN   r> Tcp !  
;
: BEGIN ( -- sca caddr # )   ?comp <mark #begin ;

: WHILE ( sca1 caddr1 #1 -- sca2 caddr2 sca1 caddr1 #2 )
   ?comp #begin ?pairs  T IF H drop 2swap #while
;
: REPEAT ( sca caddr 3 | sca2 caddr2 sca1 caddr1 4 -- )
   ?comp -rot  #branch <resolve T branch H
   #begin case? ?EXIT
   #while ?pairs >resolve ?dup IF  (if  THEN
;
: UNTIL  ( sca caddr # -- )
   ?comp  #begin ?pairs
\   Prefix @ #c-bit = IF  T cc H #nc-bit Prefix !  THEN   \ handles carry UNTIL case automatically, but changes carry!!
   Prefix @ <resolve conditional,
;

WITH_TOR [IF]

: FOR    ( -- sca caddr #for )                 ?comp T >r H <mark #for ;

: ?FOR   ( -- sca2 caddr2 sca1 addr1 #?for )   ?comp T >r H 1 (else H drop <mark 2swap #?for ;

: NEXT   ( sca caddr #for | sca2 caddr2 sca1 caddr1 #?for -- )
   ?comp #for case? IF  #nz-bit <resolve T nz-branch H EXIT THEN
        #?for ?pairs    #if T THEN H #nz-bit <resolve T nz-branch H
;

[ELSE] \ without TOR register

: FOR    ( -- sca caddr #for )
   ?comp T 1+ H <mark T 1- >r H #for
;
: ?FOR   ( -- sca1 caddr1 sca2 caddr2 #?for )
   ?comp Prefix @ #if-bit - abort" invalid Prefix"
   #z-bit Prefix ! T IF H drop <mark T 1- >r H #?for
;
: NEXT   ( sca caddr #for | sca2 caddr2 sca1 caddr1 #?for -- )
   ?comp #for case? IF  T r> H #nz-bit <resolve T nz-branch drop H EXIT THEN
   #?for ?pairs T r> H #nz-bit <resolve T nz-branch H #if T THEN drop H
;

[THEN]

: ?EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp Prefix @   no_prefix
   #if-bit  case? IF  T    nz-exit  H EXIT THEN
   #nif-bit case? IF  T     z-exit  H EXIT THEN
   #s-bit   case? IF  T    ns-exit  H EXIT THEN
   #ns-bit  case? IF  T     s-exit  H EXIT THEN
   #c-bit   case? IF  T    nc-exit  H EXIT THEN
   #nc-bit  case? IF  T cc nc-exit  H EXIT THEN
   #o-bit   case? IF  T    no-exit  H EXIT THEN
   ?strange_prefix
;
: 0=EXIT  ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )   T not ?EXIT H ;

: ?CALL  ( <name> -- ) ( zero: n -- ) ( sign: n -- n ) ( carry: ( -- )
   ?comp ' >body @ >r  Prefix @   no_prefix
   #if-bit  case? IF         r> dup #if-bit  <resolve T drop_flag nz-call  H EXIT THEN
   #nif-bit case? IF         r> dup #nif-bit <resolve T drop_flag  z-call  H EXIT THEN
   #s-bit   case? IF         r> dup #s-bit   <resolve T ns-call            H EXIT THEN
   #ns-bit  case? IF         r> dup #ns-bit  <resolve T  s-call            H EXIT THEN
   #c-bit   case? IF         r> dup #c-bit   <resolve T nc-call            H EXIT THEN
   #nc-bit  case? IF  T cc H r> dup #nc-bit  <resolve T nc-call            H EXIT THEN
   ?strange_prefix
;
: 0=CALL  ( <name> -- ) ( zero: n -- ) ( sign: n -- n ) ( carry: -- )  T not ?CALL H ;

\
\ data types in data RAM
\
Target definitions  Forth

: code-origin ( caddr -- ) ?exec Tcp ! | ;

: data-origin ( taddr -- ) ?exec Tdp !  ;
: allot       ( n -- )     Tdp +! ;

: Constant    ( n -- )     Constant immediate
                           here  Constants @ , Constants !
Does>         ( -- n )     @ comp? 0=EXIT  T Literal H
;
: Create      ( -- )       Tdp @ Constant immediate
                           here  Variables @ , Variables !
Does>         ( -- n )     @ comp? 0=EXIT  T Literal H   \ compiles a data memory address
;
: Variable    ( -- )       T Create H 1 Tdp +! ;

: Label       ( -- )       there Constant 
                           here   Labels @ , Labels ! 
Does>         ( -- n )     @ comp? 0=EXIT  T Literal H   \ returns a code memory address
; immediate

: Taskvariable ( -- )      TaskVariables @  T Constant H  1 TaskVariables +! ;

: Task        ( -- )       T Create H  TaskVariables @ Tdp +! ;

: Semaphor    ( -- )       T Create H  3 Tdp +! ; 

: Version     ( f -- )     Constant immediate ;

: [           ( -- )       ?comp postpone [ ; immediate

: ]           ( -- )       ?exec ] ;

: Code        ( -- #code ) there #smallest or Constant T ] H #code
                           here   Codes @ , Codes !
Does>         ( -- n )     @ comp? 0=EXIT  T Literal H   \ returns a code memory address
;
: end-code    ( #code -- ) #code ?pairs T ip-branch postpone [ H
; immediate

: trap-vector ( n -- addr ) ?exec trap-vector ;
\
\ Words which are useful to execute interactively during target compilation.
\ To be extended when the need arises
\
: .        ?exec . ;
: .s       ?exec .s ;
: u.       ?exec u. ;
: (              postpone ( ; immediate  \ )
: \              postpone \ ; immediate
: decimal  ?exec decimal ;
: hex      ?exec hex ;
: include  ?exec include ;
: 2**      ?exec 2** ;

\ undebugged : postpone ( -- )  ?comp postpone postpone ; immediate
\
\ setting the search order for Target compilation
\
Forth definitions

: ;USR  ( there trapaddr # -- )
   ?comp #trap ?pairs  postpone [
   there swap - 1 trap-vector u> IF  ." expanded trap"  THEN
   Tcp !
;

Target definitions Forth

: USR:  ( <name> n -- there trapaddr # )
   dup #usrmask > abort" USR number too large"
   ?exec Create dup #usr or ,
   trap-vector  there  over Tcp !
   swap ] #trap |                       \ leaves there, addr and #trap for check by ;USR
   here  Operators @ , Operators !
DOES> ( -- )  ?comp @ t, ;

: ;USR  ( there trapaddr # -- )  ;USR ; immediate

: :     ( <name> -- # )
        ?exec Create there , ]  #colon  \ leaves #colon during compilation for ";" to recognize
        here Colons @ , Colons !  |  no_prefix
Does>       ( -- )     ?comp @ dup #call <resolve T call H
;

: ;    ( # -- )
   ?comp #colon case? IF  T exit H postpone [   EXIT THEN
     dup #trap      = IF  T exit H ;USR         EXIT THEN
     dup #code      = IF  T postpone end-code H EXIT THEN
   postpone ;
; immediate

: '    ( <name> -- caddr )   ?exec ' >body @ ;

: [']  ( <name> -- caddr )   ?comp ' >body @ lit, ;

: [char] ( <c> -- )          ?comp bl parse drop c@ T literal H ; 

: ,    ( n -- )              ?comp Tdp @ lit, T ! 1 allot H ;

: here ( -- addr )           Tdp @ comp? IF  lit,  THEN ;

: immediate  ( -- )          ?exec immediate ;

Forth Root definitions

' order Alias order
' words Alias words
' \     Alias -- immediate  -- VHDL compatibility mode

have _WIN32FOR [IF]
    
: Host    ( -- )
   ['] _NUMBER,     IS NUMBER,                    \ restore "normal" INTERPRET behaviour
   ['] APP-COMPILE, IS COMPILE,
   Only Forth also definitions 
;
[THEN]

have _GFORTH [IF]

: Host
    ['] interpreter  theInterpreter !
    ['] compiler     theCompiler !
    Only Forth also definitions
    ['] interpreter is parser
;
[THEN]

: t'  ( <name> -- caddr )    Target defined Host 0= ?missing >body @ ;

' [IF]   Alias [IF] immediate
: [NOTIF]  ( f -- )
   0= postpone [IF] ; immediate
' [ELSE] Alias [ELSE] immediate
' [THEN] Alias [THEN] immediate

Forth definitions

have _WIN32FOR [IF]
    
: t_NUMBER,      ( d -- )
   DOUBLE? 0= IF DROP THEN  lit, ;

: target-compile  ( -- )
   ['] t_NUMBER, IS NUMBER,                       \ make INTERPRET compile numbers into the target memory
   ['] execute   IS COMPILE,                      \ execute all words during target compilation even in compile mode
   Only Target also definitions
;

[THEN]

have _GFORTH [IF]

: target-compiler ( addr len -- )
    \ cr ." target-compiler: " order ." | " .s ." | " 2dup type space  ." Tcp=" Tcp @ .  
    2dup find-name dup 
    IF     nip nip name>int execute 
    ELSE   drop 2dup 2>r snumber? dup
	IF  2rdrop   0>  IF   swap lit,  THEN    lit, 
	ELSE drop 2r> compiler-notfound 
	THEN 
    THEN ;

: target-compile  ( -- )
   ['] target-compiler  dup theInterpreter !  theCompiler !
   Only Target also definitions
;
[THEN]

: new      ( -- )                                 \ Initialize cross-compiler for another compilation run
   ." MicroCore cross-compiler " .version ." vom " .date
   ." by KS, gforth port by uho" cr
   Memory [ #progmask #maxprog umin 1+ cells ] Literal erase
   Tcp off  Tdp off  no_prefix  Colons off   Sequential off    Labels off
   Macro off   [ Macro# @ ] Literal Macro# !   [ Macros @ ] Literal Macros !
   [ Operators @ ] Literal Operators !
   [ Constants @ ] Literal Constants !
   [ Variables @ ] Literal Variables !
   [ Codes @     ] Literal Codes !
   [ TaskVariables @ ] Literal TaskVariables !
;

Target definitions Forth

: Macros   ( -- )  ?exec Host Target definitions ; \ This allows macro definitions during target compilation
                                                   \ switch back to target compilation with TARGET
: Compiler ( -- )  T Macros H Forth also ;

Forth definitions

: Target   ( -- )  target-compile ;

: all-opcodes  ( -- )            $80 0 DO I t, LOOP ;

: Bin-file   ( <name> -- )                       \ write program memory image to file <name> as binary
   BL word dup c@ 0= abort" file name required"
   count R/W BIN           Create-file abort" File could not be created"  >r
   0 >memory there cells r@ write-file abort" File write error"
   r>                       close-file abort" File close failed"
;

: Byte-file   ( <name> -- )                      \ write program memory image to file <name> as byte oriented binary
   BL word dup c@ 0= abort" file name required"
   count R/W BIN           Create-file abort" File could not be created"  
   there 0 ?DO I >memory @  pad c!  dup   pad 1 rot write-file abort" File write error"  LOOP
   close-file abort" File close failed"
;
: write-alias ( addr1 addr2 len file -- )
     >r rot
     base @ >r hex   0 <# #S #> r> base !  
     r@ write-file abort" File write error"
     s"  Alias "  r@ write-file abort" File write error"
     r> write-line abort" File write error"
;
: write-constant ( addr1 addr2 len file -- )
     >r rot
     base @ >r hex  dup dup 0< IF negate THEN 0 <# #s rot sign #> r> base !  
     r@ write-file abort" File write error"
     s"  Constant "  r@ write-file abort" File write error"
     r> write-line abort" File write error"
;
: write-variable ( addr1 addr2 len file -- )
     >r rot
     base @ >r hex  dup dup 0< IF negate THEN 0 <# #s rot sign #> r> base !  
     r@ write-file abort" File write error"
     s"  Variable "  r@ write-file abort" File write error"
     r> write-line abort" File write error"
;
: next_macro  ( caddr len -- caddr' len' macroaddr macrolen nameaddr namelen | caddr' ff )
   BEGIN  dup 0=EXIT            \ string exhausted?
          over macro@ 0=
   WHILE  swap 1+ swap 1-
   REPEAT                       \ macro has been found
   over macro@ >r over >r  1 >r
   BEGIN  swap 1+ swap 1-       \ advance pointer
          over macro@
          r> r> r@ -rot >r >r
          -   over 0= or   0=   \ same macro or empty string?
   WHILE  r> 1+ >r
   REPEAT
   r> r> swap    ( R: macro# )
   Macros BEGIN  @ dup 0= abort" Macro search error"
                 dup cell- @ r@ =
          UNTIL  rdrop
   2 cells - body> >name name>string  ( caddr len macroaddr macrolen nameaddr namelen )
;
: write-macros  ( file -- )
   base @ >r hex   >r   0 there
   BEGIN  next_macro ?dup
   WHILE  2swap swap 0 <# #S #> r@ write-file abort" File write error"
                     0 <# #S BL hold #> r@ write-file abort" File write error"
          s"  Macro "  r@ write-file abort" File write error"
          r@ write-line abort" File write error"
   REPEAT  drop rdrop  r> base !
;
: Sym-file ( <name> -- )
   BL word dup c@ 0= abort" file name required"
   count R/W BIN  Create-file abort" File could not be created"   >r
   S" \ MicroCore cross-compiler symbol table" r@ write-line abort" File write error"

   Colons
   BEGIN @ ?dup WHILE ( addr -- )
     dup cell- @    over cell- body> >name name>string r@ write-alias
   REPEAT

   Labels
   BEGIN @ ?dup WHILE ( addr -- )
     dup cell- @    over cell- body> >name name>string r@ write-alias
   REPEAT

   Operators
   BEGIN @ ?dup WHILE ( addr -- )
     dup cell- @  dup #usr and #usr = IF 
        #usr invert and dup #usrmask > IF  
	  drop 
	ELSE  trap-vector  over cell- body> >name name>string r@ write-alias  THEN
     ELSE drop THEN 
   REPEAT

   Constants
   BEGIN @ ?dup WHILE ( addr -- )
     dup cell- @    over cell- body> >name name>string r@ write-constant
   REPEAT

   Variables
   BEGIN @ ?dup WHILE ( addr -- )
     dup cell- @    over cell- body> >name name>string r@ write-variable
   REPEAT

   r@ write-macros

   r> close-file abort" File close failed"
;

$0A Constant #lf
$0D Constant #cr
  0 Constant #spaces

Create clause ," _00000000_,  "

: >clause  ( offset -- addr )  clause + #spaces + ;

char " $1 >clause c!
char " $A >clause c!
   #cr $C >clause c!
   #lf $D >clause c!

$2 Constant vhdlcode

: makeopcode ( opcode -- string )  
   Base @ >r  binary  0 <# # # # # # # # # #>  r> Base !
;
: make_initstring  ( caddr -- )
   >memory @ makeopcode vhdlcode >clause swap cmove
;
: writeclause  ( handle -- )
   Clause count rot write-file abort" File write error"
;
Create prolog ," prolog.vhd"
Create epilog ," epilog.vhd"

: file>here  ( string -- addr len )
   count r/o open-file abort" PROLOG/EPILOG file not present" >r
   here 2000 r@ read-file abort" PROLOG/EPILOG: could not read"
   here swap   r> close-file drop
;
: Vhdl-file  ( <name> -- )
   BL word dup c@ 0= abort" file name required"
   count R/W BIN create-file abort" File could not be created" >r
   prolog file>here r@ write-file abort" PROLOG write error"
   r> 0 there bounds  DO  I make_initstring  dup writeclause  LOOP  >r
   epilog file>here r@ write-file abort" EPILOG write error"
   r> close-file abort" File close failed" ;

